/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.Attribute;
import javax.management.AttributeChangeNotification;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationEmitter;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;

/**
 * A runtime configuration MBean which sends notifications to subscribers. It is
 * used to change configuration values at runtime.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class RuntimeConfigurationNotifier implements DynamicMBean, NotificationEmitter {

    private NotificationBroadcasterSupport broadcasterSupport;

    private ObjectName mbeanName;

    private AtomicLong seqNumber;

    private Logger logger;
    
    private Jbi jbiComponentConfiguration;

    private File jbiComponentFile;

    /**
     * 
     * @param workspaceRoot
     * @param properties
     * @throws JBIException
     */
    public RuntimeConfigurationNotifier(Jbi jbiComponentConfiguration, File jbiComponentFile, Logger logger) {
        this.jbiComponentConfiguration = jbiComponentConfiguration;
        this.jbiComponentFile = jbiComponentFile;
        this.broadcasterSupport = new NotificationBroadcasterSupport();
        this.seqNumber = new AtomicLong(0);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttribute(java.lang.String)
     */
    public synchronized Object getAttribute(String attribute) throws AttributeNotFoundException,
            MBeanException, ReflectionException {
        Object attributeObject = null;

        try {
            Field field = Component.class.getDeclaredField(attribute);
            if (field != null) {
                field.setAccessible(true);
                Object settableObject = field.get(this.jbiComponentConfiguration.getComponent());
                if (settableObject != null) {
                    String methodName = "getValue";
                    if (field.getType().getSimpleName().endsWith("boolean")) {
                        methodName = "isValue";
                    }
                    Method method = settableObject.getClass().getMethod(methodName);
                    attributeObject = method.invoke(settableObject);
                }
            } else {
                throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
            }
        } catch (SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' is not reachable");
        } catch (NoSuchFieldException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
        } catch (IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (NoSuchMethodException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        return attributeObject;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttributes(java.lang.String[])
     */
    public synchronized AttributeList getAttributes(String[] attributes) {
        AttributeList list = new AttributeList();

        for (String attribute : attributes) {
            try {
                list.add(new Attribute(attribute, this.getAttribute(attribute)));
            } catch (AttributeNotFoundException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (MBeanException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (ReflectionException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            }
        }

        return list;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttribute(javax.management.Attribute)
     */
    public synchronized void setAttribute(Attribute attribute) throws AttributeNotFoundException,
            InvalidAttributeValueException, MBeanException, ReflectionException {

        try {
            Field field = Component.class.getDeclaredField(attribute.getName());
            if (field != null) {
                field.setAccessible(true);
                Object settableObject = field.get(this.jbiComponentConfiguration.getComponent());
                // get the previous value
                String methodName = "getValue";
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    methodName = "isValue";
                }
                Method method = settableObject.getClass().getMethod(methodName);
                Object value = method.invoke(settableObject);

                // set the value
                Field primitiveField = field.getType().getDeclaredField("value");
                method = settableObject.getClass().getMethod("setValue",
                        new Class[] { primitiveField.getType() });
                method.invoke(settableObject, new Object[] { attribute.getValue() });

                // notify the listeners about the value modification
                saveAndNotifyListeners(attribute.getName() + "modified", attribute.getName(),
                        value, attribute.getValue());
            } else {
                throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                        + "' does not exist");
            }
        } catch (SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' is not reachable");
        } catch (NoSuchFieldException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' does not exist");
        } catch (IllegalArgumentException e) {
            throw new ReflectionException(e);
        } catch (IllegalAccessException e) {
            throw new ReflectionException(e);
        } catch (NoSuchMethodException e) {
            throw new ReflectionException(e);
        } catch (InvocationTargetException e) {
            throw new ReflectionException(e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttributes(javax.management.AttributeList
     * )
     */
    public synchronized AttributeList setAttributes(final AttributeList list) {
        final AttributeList attributeList = new AttributeList();

        for (Object object : list) {
            final Attribute attribute = (Attribute)object;
            try {
                Field field = Component.class.getDeclaredField(attribute.getName());
                if (field != null) {
                    field.setAccessible(true);
                    final Object settableObject = field
                            .get(this.jbiComponentConfiguration.getComponent());
                    final Field primitiveField = field.getType().getDeclaredField("value");
                    final Method method = settableObject.getClass().getMethod("setValue",
                            new Class[] { primitiveField.getType() });
                    method.invoke(settableObject,
                            new Object[] { attribute.getValue() });

                    attributeList.add(attribute);

                } else {
                    throw new AttributeNotFoundException("Attribute '"
                            + attribute.getName() + "' does not exist");
                }
            } catch (final Exception e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            }
        }

        return attributeList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        MBeanAttributeInfo[] attrs = getAttributesInfos();

        MBeanNotificationInfo notif = new MBeanNotificationInfo(null, "Runtime MBean notifier",
                "Runtime configuration modification notifier");
        MBeanNotificationInfo[] notifs = new MBeanNotificationInfo[] { notif };

        // create MBeanInfo with no constructors and no operations
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null,
                null, notifs);
    }

    public Object invoke(String actionName, Object[] params, String[] signature)
            throws MBeanException, ReflectionException {
        // TODO Auto-generated method stub
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] { new MBeanNotificationInfo(
                new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE },
                AttributeChangeNotification.class.getName(),
                "Change runtime configuration attribute") };
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#addNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void addNotificationListener(NotificationListener listener, NotificationFilter filter,
            Object handback) {
        this.broadcasterSupport.addNotificationListener(listener, filter, handback);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#removeNotificationListener(javax
     * .management.NotificationListener)
     */
    public void removeNotificationListener(NotificationListener listener)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationEmitter#removeNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void removeNotificationListener(NotificationListener listener,
            NotificationFilter filter, Object handback) throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener, filter, handback);
    }

    /**
     * Get the MBean Name
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return this.mbeanName;
    }

    /**
     * 
     * @param mbeanName
     */
    public void setMBeanName(ObjectName mbeanName) {
        this.mbeanName = mbeanName;
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     * 
     * @return
     */
    protected MBeanAttributeInfo[] getAttributesInfos() {
        MBeanAttributeInfo[] attrs = null;
        Field[] fields = Component.class.getDeclaredFields();
        List<Field> settableFields = new ArrayList<Field>();

        try {
            // get the settable fields
            for (Field field : fields) {
                if (field.getType().getSimpleName().startsWith("Runtime")) {
                    field.setAccessible(true);
                    if (field.get(this.jbiComponentConfiguration.getComponent()) != null) {
                        settableFields.add(field);
                    }
                }
            }

            // build the MBeanAttributeInfo
            attrs = new MBeanAttributeInfo[settableFields.size()];
            for (Field field : settableFields) {
                Object settableObject = field.get(this.jbiComponentConfiguration.getComponent());
                Method getter;
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    getter = settableObject.getClass().getMethod("isValue");
                } else {
                    getter = settableObject.getClass().getMethod("getValue");
                }
                Field primitiveField = field.getType().getDeclaredField("value");
                Method setter = settableObject.getClass().getMethod("setValue",
                        new Class[] { primitiveField.getType() });

                attrs[settableFields.indexOf(field)] = new MBeanAttributeInfo(field.getName(),
                        "Property " + field.getName(), getter, setter);

            }

        } catch (IllegalArgumentException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (IllegalAccessException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (SecurityException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (NoSuchMethodException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (IntrospectionException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (NoSuchFieldException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        return attrs;
    }

    /**
     * FIXME : New value to be saved in the properties file.
     * 
     * @param msg
     * @param attrName
     * @param oldVal
     * @param newVal
     */
    private void saveAndNotifyListeners(String msg, String attrName, Object oldVal, Object newVal) {

        try {
            JBIDescriptorBuilder.buildXmlJBIdescriptor(this.jbiComponentConfiguration,
                    new FileOutputStream(jbiComponentFile));
        } catch (FileNotFoundException e) {
            this.logger
                    .warning("Failed to pesist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (CDKJBIDescriptorException e) {
            this.logger
                    .warning("Failed to pesist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        Notification notif = new AttributeChangeNotification(this, seqNumber.getAndIncrement(),
                System.currentTimeMillis(), msg, attrName, String.class.getName(), oldVal, newVal);
        this.broadcasterSupport.sendNotification(notif);
    }
}
